#!/usr/bin/env python
# coding: utf-8

# In[ ]:


class Dataframe_extraction_preprocessing:
    """ Class used to the extraction of data from .csv files
        The extraction includes setting the time, handling of missing values and
        removing turbines which does not contain measurements.
        dataset_name us required because for one data set the filled values are wrongly filled."""
    def __init__(self, path_csv, filenames, dataset_name, skip_rows):
        self.dataframes = []
        self.dataframes_norm = []
        self.dataframes_scaled = []
        
        # Importing into dataframes
        for idx, filename in enumerate(filenames):
            self.dataframes.append(pd.read_csv(path_csv+filename+'.csv', delimiter=',', skiprows = skip_rows[idx]))
         
        self.setting_time_dataframe()       # Setting time for dataframes   
        self.filling_missing_values(dataset_name)        # Handling missing values
        self.remove_turbines_with_no_measurement() 
    
    def setting_time_dataframe(self):
        """ Setting date time for the dataframe if not already done
            Also renames the columns to the correct turbine number """
        for idx, df in enumerate(self.dataframes):
            try:
                df['Time'] = pd.to_datetime(df['Time'], format='%Y-%m-%dT%H:%M:%S.%fZ')
                df.set_index('Time', inplace=True)
                df.columns = ['WT1', 'WT2', 'WT3','WT5', 'WT6', 'WT7', 'WT8', 'WT9', 
                              'WT10', 'WT11', 'WT12', 'WT13', 'WT14', 'WT15', 'WT16']                
            except:
                try:
                    df.columns = ['WT1', 'WT2', 'WT3', 'WT4', 'WT5', 'WT6', 'WT7', 'WT8', 'WT9', 
                                  'WT10', 'WT11', 'WT12', 'WT13', 'WT14', 'WT15', 'WT16']
                except: 
                    print("Dataframe {} does not fit column names".format(idx+1))
                    print(df.columns)
                    
    
    def filling_missing_values(self, dataset_name):
        # The function fills in wrong for this timeseries associated with 'Mulitvariate_v2' data set
        if dataset_name == 'Multivariate_v2':
            self.dataframes[0]['WT2'].fillna(0, inplace=True)
        for df in self.dataframes:
            number_of_NaN_before = df.isna().sum().sum()
            df.fillna(method='ffill', inplace=True) # Filling inn all NaN with forward filling
            df.fillna(method='bfill', inplace=True) # In case the first values are NaN, do backwards filling.
            number_of_NaN_after = df.isna().sum().sum()
            print("Number of NA fields in dataset", number_of_NaN_before)
            print("Number of NA fields in dataset after filling: ", number_of_NaN_after,"/",number_of_NaN_before)

            
    def remove_turbines_with_no_measurement(self):
        """ Removes turbine with no measurments
            and corrects the shape of the timeseries"""
        # Finding the turbine which contains missing values
        list_df = []
        for df in self.dataframes:
            number_of_NaN = df.isna().sum()
            for idx in np.where(number_of_NaN > 0)[0]:
                if df.columns[idx][:6] not in list_df: 
                    list_df.append(df.columns[idx][:6])

        # Removing the column containing nan from all dataframes
        for idx, df in enumerate(self.dataframes):
            for index_name in list_df:
                try:
                    df.drop([index_name+df.columns[0][6:]], axis=1, inplace=True)
                except:
                    print(index_name+df.columns[0][6:]+" Does not exist in df", idx + 1)
        
        # Fixing timestamps such that they are equal
        self.print_shape_time_interval()
        dataframes_new = []
        for idx, df in enumerate(self.dataframes):
            dataframes_new.append(df[:self.dataframes[0].shape[0]])        
        self.dataframes = dataframes_new
        self.print_shape_time_interval()
        
        
    def print_shape_time_interval(self):
        for idx, df in enumerate(self.dataframes):
            print("Shape of dataframe{0}: {1},".format(idx+1, df.shape), end='        ')
            print("Start/stop", df.index[0] ,"/", df.index[-1])  
    

    def running_median_insort(self,seq, window_size): # Faster approach to simple moving median filter
        """Contributed by Peter Otten"""
        seq = iter(seq)
        d = deque()
        s = []
        result = []
        for item in islice(seq, window_size):
            d.append(item)
            insort(s, item)
            result.append(s[len(d)//2])
        m = window_size // 2
        for item in seq:
            old = d.popleft()
            d.append(item)
            del s[bisect_left(s, old)]
            insort(s, item)
            result.append(s[m])
        return np.asarray(result)

    
    # Filters the dataset in the object with a preent lenght of the moving window
    def filter_median(self, length_moving_median):
        for df in self.dataframes:
            for i in range(df.shape[1]):
                df.values[:,i] = self.running_median_insort(df.values[:,i], length_moving_median)

    
    def normalization(self):    
        for df in self.dataframes:
            self.dataframes_norm.append((df - np.min(df.values, axis=0)) / (np.max(df.values, axis=0) - 
                                                                            np.min(df.values, axis=0)))        
    
    
    def scaling(self):
        for df in self.dataframes:
            temp = (df - np.min(np.min(df, axis=0)))
            temp /= np.max(np.max(temp, axis=0))
            self.dataframes_scaled.append(temp)
    
    # Helper function for plotting the different data sets
    # initial, filtered, scaled, normalised data set
    def plotting_dataframe(self, dataframe, title="No title"):
            style_code = ['-', '-', '-', '-', '-', '-', '-', '-', '-', '-',
                  '--','--','--','--','--','--','--','--','--','--',]
            print("\n",title,"\n")
            for df in dataframe:
                df.plot(figsize=(18,10), subplots=True, layout=(4,4), sharex=True, sharey=True, grid=False, 
                        style=style_code, x_compat=True, fontsize= 10)
                plt.gca()
                plt.show()

        

