#!/usr/bin/env python
# coding: utf-8

# In[ ]:


class Internal_indexes:
    """ Class for calculating and plotting of the internal indexes. 
        dist_cond: 1D condensed distance matrix calculated with the the weighting vector, as used in SciPy.
        dist_cond_original: 1D condensed distance matrix calculated without the weighting vector (all alphas equal to 1)
        count_nonzero: Number of nonzero elements in Alpha vector.
        
        Procedure for calculating and presenting the itnernal indexes
        1) First initate the class witht du corresponding distance matrices, count_nonzeroes and dataframes_scaled
           for the euclidean distance or random dataframe for the dtw distance
        2) Call the 'calculation_internal_indexes' with the list of methods allowed in the SciPy notation and valid metric
           'euclidean' or 'dtw'.
        3) Call 'plotting_internal_indexes' for potting of the internal indexs
        """
    
    def __init__(self, dist_cond, dist_cond_original, count_nonzero, dataframes):
        self.dist_cond = dist_cond
        self.dist_cond_original = dist_cond_original
        self.count_nonzero = count_nonzero
        self.dataframes = dataframes
        self.correlations = {}
        self.silhouettes = {}
        self.mean_sum_squared = {}
    
    def cophenet_index(self, Z):
        """Calculates the cophenet correlation coefficent for a given dendrogram, Z"""
        c = cophenet(Z)
        corr_value = np.corrcoef(c,self.dist_cond)[0,1]
        return corr_value
    
    
    
    def silhouette_index(self, Z):
        """ Calculates the silhouette index for a given dendrogram, Z
            for cuts from 2 to number of (turbines - 2) """
        dist_squared = squareform(self.dist_cond)
        silhouette_scores = []
        for k in range(2,dist_squared.shape[0] - 1):
            labels = fcluster(Z, k, criterion='maxclust')
            silhouette_scores.append(silhouette_score(X=dist_squared, labels=labels, metric="precomputed"))
        return silhouette_scores
    

    # Helper function for sum_squared_error_euc and sum_squared_error_dtw functions
    def mssse(self, columns, SSE):
        """ Calculates the average sum of SSE for all clusters formed. """
        MSSSE = 0
        for index_name in columns:
            MSSSE += SSE[index_name]
        MSSSE /= len(columns)
        return MSSSE

    
    
    def sum_squared_error_euc(self, Z, current_scaled_dataframes):
        dist_squared_list = []
        # Converting list of 1d condensed matrices to square form
        for idx, element in enumerate(self.dist_cond_original):
            dist_squared_list.append(squareform(element))

        mean_sum_of_SSE = []
        for num_cuts in range(2,dist_squared_list[0].shape[0] - 1):
            # Getting labels
            labels = fcluster(Z, num_cuts, criterion='maxclust')
            s = pd.Series(labels)
            clusters = s.unique()
            clusters.sort()

            cluster_vectors = pd.DataFrame()
            sum_squared_error = {}  # dict for storing the SSE
            for c in clusters:
                cluster_indeces = s[s==c].index
                # Caluclating the mean vector for df in scaled_data = [df1, df2, df3, ...]
                for idx, df in enumerate(current_scaled_dataframes):
                    cluster_vectors['Cluster{0}_{1}'.format(c,idx)] = df[df.columns[cluster_indeces]].sum(axis=1)/
                                                                                            len(cluster_indeces)

                # Getting the sum of squared error between all timeseries in the cluster to its cluster centre
                # This is repeated for each dimension scaled_data = [df1, df2, df3, ...] and than the mean 
                # is extracted
                squared_error = {}
                for idx, df in enumerate(current_scaled_dataframes): # For [df1,df2,df3]
                    temp = 0
                    for timeseries in cluster_indeces: #Indces for the timeseres within the current cluster
                        temp2 = np.sum( (df[df.columns[timeseries]] -
                                         cluster_vectors['Cluster{0}_{1}'.format(c,idx)])**2 )
                        squared_error[df.columns[timeseries]+'_cluster_'+str(idx)] = temp2

                # Initializing each sum to zero
                for timeseries in cluster_indeces:  # indces for the timeseres within the current cluster
                    sum_squared_error[df.columns[timeseries]] = 0

                # Summing each squared error associated with the same turbine (f.eks WT2)
                # from all dimension in scaled_dataframes: df1, df2, df3, ...
                for idx, df in enumerate(current_scaled_dataframes): # For [df1,df2,df3]
                    for timeseries in cluster_indeces:
                        sum_squared_error[df.columns[timeseries]] += squared_error[df.columns[timeseries]+
                                                                                   '_cluster_'+str(idx)]


            # Dividing all errors by the number of dimension to get a comparable value
            for idx in range(len(sum_squared_error)):
                sum_squared_error[df.columns[idx]] /= len(current_scaled_dataframes)


            #Mean sum of SSE - as above, to get a comparable value
            temp = self.mssse(current_scaled_dataframes[0].columns, sum_squared_error)
            mean_sum_of_SSE.append(temp)

        return mean_sum_of_SSE
    

    def get_centroid(self, cluster_indeces, dist_squared):
        # Cluster_indeces is all timeseries in that cluster with index values. Same as the function returns.
        centroid_distances = []
        for ts in cluster_indeces:
            average_distance = dist_squared[ts,cluster_indeces.drop(ts)].sum(axis=0)
            centroid_distances.append(average_distance)

        # Returns index from [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14]
        # Indexes associated ['WT1','WT2','WT3','WT5','WT6','WT7','WT8','WT9','WT10','WT11','WT12','WT13',
        # 'WT14','WT15','WT16']
        centroid_index = cluster_indeces[centroid_distances.index(min(centroid_distances))]
        return centroid_index



    def sum_squared_error_dtw(self, Z, columns):
        dist_squared = squareform(self.dist_cond)
        mean_sum_of_SSE = []
        for num_cuts in range(2,dist_squared.shape[0] - 1):
            # Getting labels
            labels = fcluster(Z, num_cuts, criterion='maxclust')
            s = pd.Series(labels)
            clusters = s.unique()
            clusters.sort()

            centroid_indices = []
            sum_squared_error = {}  # dict for storing the SSE
            for c in clusters:
                cluster_indeces = s[s==c].index

                # Finding the centroid for each cluster
                centroid_index = self.get_centroid(cluster_indeces, dist_squared)
                centroid_indices.append(centroid_index)

                # Calculating the sum of squared error for each timeseries to its centroid
                for ts1 in cluster_indeces:
                    sum_squared_error[columns[ts1]] = dist_squared[centroid_index, ts1]
                    #print(centroid_index, ts1, dist_squared[centroid_index, ts1])

            # Overall Mean sum of SSE
            temp = self.mssse(columns, sum_squared_error)
            mean_sum_of_SSE.append(temp)

        return mean_sum_of_SSE 


    
    def calculation_internal_indexes(self, methods, metric, similarity_measure = "euclidean"):
        """ metric and methods limited to those avaiable in the SciPy hac library.
            methods: should be a list containing the names of linkage criterias. 
                    Example: methods = ['single', 'complete'] or methods = ['single']
            metric: internal distance, measure used in the linakge criteria
            similarity_measure: can either be the 'euclidean' or the 'dtw'"""
        for method in methods:
            Z = hac.linkage(self.dist_cond, method=method, metric=metric, optimal_ordering=True)
            self.correlations[method] = self.cophenet_index(Z)
            self.silhouettes[method] = self.silhouette_index(Z)
            if similarity_measure == "euclidean":
                self.mean_sum_squared[method] = self.sum_squared_error_euc(Z, self.dataframes[:self.count_nonzero])
            elif similarity_measure == "dtw":
                self.mean_sum_squared[method] = self.sum_squared_error_dtw(Z, self.dataframes[0].columns)
            else:
                print("Invalid similarity measure. Should be either 'euclidean' or 'dtw'. MSSSE is not calculated")

            print("Method: "+method)
            print("Correlation: {:.3f}".format(self.correlations[method]))
            print("Silhouettes:   \t\t", end='')
            print(' [' + ';'.join('%6.2f' % v for v in self.silhouettes[method]) + ']')
            if self.mean_sum_squared:
                print("Mean sum of SSE: \t", end='')
                print(' [' + ';'.join('%6.2f' % v for v in self.mean_sum_squared[method]) + ']\n')
            
        for method in methods:
            print(';'.join('%6.2f' % v for v in self.silhouettes[method]))
            if self.mean_sum_squared:
                print(';'.join('%6.2f' % v for v in self.mean_sum_squared[method]))
            
            

    def plotting_internal_indexes(self, similarity_measure, dataset_name, cophenet_yaxis,
                                  file_ending, save_figure=False):
        """ similarity_measure, dataset_name and file_ending are all used to determine the
            name of the figure saved. cophenet_yaxis: number which indicates where the 
            cophenet index values should be located in the plot (y-axis)."""
        methods = ['single', 'complete', 'average', 'ward']
        fig, ax = plt.subplots(2, 2,figsize=(20,12), sharex=True, sharey=True)
        
        x_data = [] # X-axis for the coefficents
        dist_squared = squareform(self.dist_cond)
        [x_data.append(cuts) for cuts in range(2,dist_squared.shape[0] - 1)]
        temp = -1 # Used to iterate the subplots
        for idx, method in enumerate(methods):
            if idx%2 == 0:
                temp += 1
            color = 'tab:red'
            ax[temp,idx%2].grid(True)
            ax[temp,idx%2].set_title(method, fontsize=25)
            ax[temp,idx%2].plot(x_data, self.silhouettes[method],'-o', color=color, label = 'Silhouette index')
            ax[temp,idx%2].tick_params(axis='y', labelcolor=color, labelsize=15)
            ax[temp,idx%2].text(10,cophenet_yaxis,"Cophenet index: {:.3f}".format(self.correlations[method]),
                                fontsize=15)
            ax[temp,idx%2].tick_params(axis='x', labelsize=15)
            ax[temp,idx%2].set_ylabel('Silhouette index', color=color, fontsize=20)
            ax2 = ax[temp,idx%2].twinx()  # instantiate a second axes that shares the same x-axis

            color = 'tab:blue'
            ax2.plot(x_data, self.mean_sum_squared[method], '-o', color=color, label = 'Mean sum of SSE')
            ax2.tick_params(axis='y', labelcolor=color, labelsize=15)
            ax2.set_ylabel('Mean sum of SSE', color=color, fontsize=20)


        ax[1,0].set_xlabel('Number of cuts', labelpad=0, fontsize=20)
        ax[1,1].set_xlabel('Number of cuts', labelpad=0, fontsize=20)

        fig.tight_layout()  # otherwise the right y-label is slightly clipped
        plt.show()
        
        if save_figure:
            path_pictures = 'C:/Users/ewaaga/OneDrive - KONGSBERG MARITIME AS/Pictures/Performance indicators/'
            fig.savefig(path_pictures+'Internal_indices_'+similarity_measure+'_'+dataset_name+file_ending+'.eps',
                        bbox_inches='tight', pad_inches=0)

